"""
One Zero bank official webpage:
https://www.onezerobank.com/currencies/
It loads the data from:
https://dv16ymfyh91nr.cloudfront.net/MarketingRatesReport/MarketingSiteFCYRatesCurrentReport.json
"""

import pandas as pd

from xil._currencies import CurrencyCode
from xil._df_normalizer import BaseDataFrameNormalizer

_ONEZERO_URL = "\
https://dv16ymfyh91nr.cloudfront.net/MarketingRatesReport/MarketingSiteFCYRatesCurrentReport.json"
_IDX0 = pd.MultiIndex.from_product([["currency"], ["code", "official rate"]])
_IDX1 = pd.MultiIndex.from_product([["transfer"], ["buy", "sell"]])
_IDX = _IDX0.append(_IDX1)


def get_onezero_df(url: str = _ONEZERO_URL) -> pd.DataFrame:
    """Get One Zero bank exchange rates"""
    series = pd.read_json(url, typ="series")
    # series.generatingReportDateTime holds the date in YYYY-MM-DD format
    df = pd.json_normalize(series.marketingRecords)  # pylint: disable=no-member
    df = df[df.fromCurrency == CurrencyCode.ILS]  # remove data duplication
    df.drop(labels="fromCurrency", axis="columns", inplace=True)
    df.columns = _IDX
    df = BaseDataFrameNormalizer(df).norm()
    return df



"""
Jerusalem bank exchange data

https://www.bankjerusalem.co.il/capital-market/rates
"""

import pandas as pd

from xil._df_normalizer import DataFrameNormalizer

_JERUSALEM_URL = "https://www.bankjerusalem.co.il/capital-market/rates"
_HEADER = None  # the table's header is not recognized
_AMOUNT_NAME_SPLITTER = "\xa0"
_IDX0 = pd.MultiIndex.from_product(
    [["currency"], ["name", "official rate", "change (%)"]]
)
_IDX1 = pd.MultiIndex.from_product([["cash", "transfer"], ["sell", "buy"]])
_IDX = _IDX0.append(_IDX1)


class JerusalemNormalizer(DataFrameNormalizer):
    """DataFrame normalizer for Jerusalem bank data"""

    @staticmethod
    def _fix_amount(raw_amount: str) -> str:
        return raw_amount.split(_AMOUNT_NAME_SPLITTER)[1]

    @classmethod
    def _preprocess_names(cls, names: pd.Series) -> pd.Series:
        return names.apply(cls._fix_amount)


def get_jerusalem_df(url: str = _JERUSALEM_URL) -> pd.DataFrame:
    """Get Jerusalem Bank exchange rates"""
    df = pd.read_html(url, header=_HEADER)[0]
    df.columns = _IDX
    df = JerusalemNormalizer(df).norm()
    return df


if __name__ == "__main__":
    print(get_jerusalem_df())



"""
Currencies DataFrame (df) normalization.
"""

import pandas as pd

from xil._currencies import currency_from_heb_name

_CURRENCY_KEY = "currency"
_IDX0 = pd.MultiIndex.from_product(
    [[_CURRENCY_KEY], ["name", "official rate", "change (%)"]]
)
_IDX1 = pd.MultiIndex.from_product([["transfer", "cash"], ["buy", "sell"]])
_IDX = _IDX0.append(_IDX1)
_CURRENCY_CODE_KEY = (_CURRENCY_KEY, "code")
_CURRENCY_NAME_KEY = _IDX0[0]


class BaseDataFrameNormalizer:
    """
    Base class that can be used when the df has a ("currency", "code") column.
    """

    def __init__(self, df: pd.DataFrame):
        """Initialize a DataFrame (df) normalizer with a raw currencies df"""
        self.df = df

    def norm(self) -> pd.DataFrame:
        """
        Normalize the df inplace and return it - set the currency code as the index.
        """
        self.set_code_index()
        return self.df

    def set_code_index(self) -> None:
        """Set the ("currency", "code") as the df index"""
        self.df.set_index(_CURRENCY_CODE_KEY, inplace=True)


class DataFrameNormalizer(BaseDataFrameNormalizer):
    """
    This class is used to normalize currencies data frames of with a
    ("currency", "name") column.
    """

    def norm(self) -> pd.DataFrame:
        """Normalize the df inplace according to the given parameters and return it"""
        self.add_code_from_name()
        self.drop_currency_name()
        return super().norm()

    def add_code_from_name(self) -> None:
        """Add a ("currency", "code") column from the ("currency", "name") column"""
        self.df[_CURRENCY_CODE_KEY] = self._preprocess_names(
            self.df[_CURRENCY_NAME_KEY]
        ).apply(currency_from_heb_name)

    @classmethod
    def _preprocess_names(cls, names: pd.Series) -> pd.Series:
        """
        A preprocessing hook to manipulate the names before passing them to the
        currency_from_heb_name function.
        """
        return names

    def drop_currency_name(self) -> None:
        """Drop the ("currency", "name") column from the df"""
        self.df.drop(labels=_CURRENCY_NAME_KEY, axis="columns", inplace=True)


class JPYNormalizer(DataFrameNormalizer):
    """
    Normalizer subclass for fixing JPY amount in the name.
    Used in Poalim and Leumi banks.
    """

    _JPY_AMOUNT = "100"

    @classmethod
    def _remove_100(cls, raw_name: str) -> str:
        """
        remove '100' from '100 ין יפני', prefix or suffix.
        Note: this method does not check the currency name.
        """
        return (
            raw_name.removeprefix(cls._JPY_AMOUNT).removesuffix(cls._JPY_AMOUNT).strip()
        )

    @classmethod
    def _preprocess_names(cls, names: pd.Series) -> pd.Series:
        names = super()._preprocess_names(names)
        return names.apply(cls._remove_100)



"""
Scrape Bank Hapoalim exchange data publicly visible on
https://www.bankhapoalim.co.il/he/foreign-currency/exchange-rates
"""

from datetime import date

import pandas as pd

from xil._df_normalizer import JPYNormalizer

_POALIM_GET_URL = "https://www.bankhapoalim.co.il/he/coin-rates"
_DATE_QUERY = "?date="
_POALIM_QUERY = _POALIM_GET_URL + _DATE_QUERY
_DATE_FORMAT = "%Y-%m-%d"  # YYYY-MM-DD
_RELEVANT_COLS = [
    "DT_ERECH",
    "NAME_HEB",
    "SHAAR_YATZIG",
    "AHUZ_SHINUI",
    "SHAAR_CHK_KNIA",
    "SHAAR_CHK_MECHIRA",
    "SHAAR_CASH_KNIA",
    "SHAAR_CASH_MECHIRA",
]
_IDX0 = pd.Index(["date"])
_IDX1 = pd.MultiIndex.from_product(
    [["currency"], ["name", "official rate", "change (%)"]]
)
_IDX2 = pd.MultiIndex.from_product([["transfer", "cash"], ["buy", "sell"]])
_IDX = _IDX0.append(_IDX1.append(_IDX2))


def _get_url(t: date | None) -> str:
    if t is None:
        return _POALIM_GET_URL
    return _POALIM_QUERY + t.strftime(_DATE_FORMAT)


def get_df(t: date | None = None, keep_last_date_only: bool = True) -> pd.DataFrame:
    """
    Get poalim exchange data from the latest available day or a specified
    date t as a pandas DataFrame. If keep_last_date_only is true, and there
    is no specified date t, only the last available date's data is returned.
    """
    df = pd.read_json(_get_url(t))
    if df.empty:  # pylint: disable=no-member
        return df

    df = df[_RELEVANT_COLS]
    df.columns = _IDX

    df.date = pd.to_datetime(df.date, format=_DATE_FORMAT).dt.date

    if keep_last_date_only and t is None:
        df = df[df.date == df.date.max()]

    df = JPYNormalizer(df).norm()
    return df



"""
Utilities for currency conversion and standardization.
"""

from enum import StrEnum


class CurrencyCode(StrEnum):
    """3 letter currency codes of relevant currencies (ISO 4217)"""

    ILS = "ILS"  # Israeli New Shekel
    USD = "USD"  # US Dollar
    GBP = "GBP"  # British Pound
    JPY = "JPY"  # Japanese Yen
    EUR = "EUR"  # Euro
    AUD = "AUD"  # Australian Dollar
    CAD = "CAD"  # Canadian Dollar
    DKK = "DKK"  # Danish Krone
    NOK = "NOK"  # Norwegian Krone
    ZAR = "ZAR"  # South African Rand
    SEK = "SEK"  # Swedish Krona
    CHF = "CHF"  # Swiss Franc
    HKD = "HKD"  # Hong Kong Dollar
    SGD = "SGD"  # Singapore Dollar
    NZD = "NZD"  # New Zealand Dollar
    TRY = "TRY"  # Turkish Lira
    JOD = "JOD"  # Jordanian Dinar
    AED = "AED"  # UAE Dirham
    EGP = "EGP"  # Egyptian Pound
    ETB = "ETB"  # Ethiopian birr
    LKR = "LKR"  # Sri Lankan rupee
    NGN = "NGN"  # Nigerian naira
    KRW = "KRW"  # South Korean won
    BGN = "BGN"  # Bulgarian Lev
    CNY = "CNY"  # Yuan Renminbi
    MXN = "MXN"  # Mexican Peso
    SAR = "SAR"  # Saudi Riyal
    HUF = "HUF"  # Hungarian Forint
    INR = "INR"  # Indian Rupee
    PEN = "PEN"  # Peruvian Sol
    IDR = "IDR"  # Indonesian Rupiah
    RUB = "RUB"  # Russian Ruble
    TWD = "TWD"  # New Taiwan Dollar
    HRK = "HRK"  # Croatian Kuna
    SAL = "SAL"  # Salvadoran Colon
    THB = "THB"  # Thai Baht
    CZK = "CZK"  # Czech Koruna
    PHP = "PHP"  # Philippine Peso
    LBP = "LBP"  # Lebanese Pound
    PLN = "PLN"  # Polish Zloty
    BRL = "BRL"  # Brazilian Real
    CLP = "CLP"  # Chilean Peso
    ARS = "ARS"  # Argentine Peso
    RON = "RON"  # Romanian Leu

    # ISO 4217 non-currencies
    XAG = "XAG"  # Silver
    XAU = "XAU"  # Gold
    XPD = "XPD"  # Palladium
    XPT = "XPT"  # Platinum


_HEB_CURRENCY_NAME_TO_CODE: dict[str, CurrencyCode] = {
    "שקל חדש": CurrencyCode.ILS,
    'דולר ארה"ב': CurrencyCode.USD,
    "לירה שטרלינג": CurrencyCode.GBP,
    'ליש"ט': CurrencyCode.GBP,
    'ליש"ט בריטי': CurrencyCode.GBP,
    "יין יפני": CurrencyCode.JPY,
    "ין יפני": CurrencyCode.JPY,
    "אירו": CurrencyCode.EUR,
    "דולר אוסטרלי": CurrencyCode.AUD,
    "דולר קנדי": CurrencyCode.CAD,
    "כתרים דניים": CurrencyCode.DKK,
    "כתר דני": CurrencyCode.DKK,
    "כתר נורבגי": CurrencyCode.NOK,
    "כתר נורווגי": CurrencyCode.NOK,
    'רנד דרא"פ': CurrencyCode.ZAR,
    'ראנד דרא"פ': CurrencyCode.ZAR,
    "רנד דרום אפריקני": CurrencyCode.ZAR,
    "רנד ד. אפריקני": CurrencyCode.ZAR,
    "ראנד דרום אפריקה": CurrencyCode.ZAR,
    "כתר שוודי": CurrencyCode.SEK,
    "כתר שבדי": CurrencyCode.SEK,
    "פרנק שוויצרי": CurrencyCode.CHF,
    "פרנק שויצרי": CurrencyCode.CHF,
    "דולר הונג קונג": CurrencyCode.HKD,
    "דולר סינגפור": CurrencyCode.SGD,
    "דולר סינגפורי": CurrencyCode.SGD,
    "דולר ניו זילנד": CurrencyCode.NZD,
    "דולר ניוזלנדי": CurrencyCode.NZD,
    "דולר ניו זילנדי": CurrencyCode.NZD,
    "דולר ניוזילנד": CurrencyCode.NZD,
    "לירה טורקית": CurrencyCode.TRY,
    "לירה טורקית חדשה": CurrencyCode.TRY,
    "דינר ירדני": CurrencyCode.JOD,
    "דירהם אמירתי": CurrencyCode.AED,
    "לירה מצרית": CurrencyCode.EGP,
}


class CurrencyNotSupportedError(ValueError):
    """Raised when a currency is not supported"""


def currency_code_from_heb_name(heb_currency_name: str) -> CurrencyCode:
    """Converts a Hebrew currency name to a currency code"""
    try:
        return _HEB_CURRENCY_NAME_TO_CODE[heb_currency_name]
    except KeyError as exc:
        raise CurrencyNotSupportedError(
            f"Unknown currency name: '{heb_currency_name}'"
        ) from exc


def currency_from_heb_name(heb_currency_name: str) -> str:
    """Converts a Hebrew currency name to a currency string"""
    # noinspection PyTypeChecker
    return currency_code_from_heb_name(heb_currency_name).value


def optional_currency_from_heb_name(heb_currency_name: str) -> str | None:
    """
    Converts a Hebrew currency name to a currency string or None if not supported.
    Prefer using this function over `currency_from_heb_name` when the currency is
    not guaranteed to be supported.
    """
    try:
        # noinspection PyTypeChecker
        return currency_code_from_heb_name(heb_currency_name).value
    except CurrencyNotSupportedError:
        return None



"""
Bank of Israel (BOI) official exchange rates data.

The data is available on the following pages (when viewing the source):
https://www.boi.org.il/roles/markets/%D7%A9%D7%A2%D7%A8%D7%99-%D7%97%D7%9C%D7%99%D7%A4%D7%99%D7%9F-%D7%99%D7%A6%D7%99%D7%92%D7%99%D7%9D/
https://www.boi.org.il/en/economic-roles/financial-markets/exchange-rates/

The public API, which is also more accurate, is available with the following URL:
https://boi.org.il/PublicApi/GetExchangeRates

A slightly different URL can also be used to get a single key, e.g.:
https://boi.org.il/PublicApi/GetExchangeRate?key=USD
For more information, see:
https://www.boi.org.il/%D7%A9%D7%90%D7%9C%D7%95%D7%AA-%D7%95%D7%AA%D7%A9%D7%95%D7%91%D7%95%D7%AA-%D7%A2%D7%9C-%D7%94%D7%A9%D7%99%D7%9E%D7%95%D7%A9-%D7%91%D7%90%D7%AA%D7%A8-%D7%94%D7%97%D7%93%D7%A9/
"""

from functools import partial

import pandas as pd

from xil._df_normalizer import _CURRENCY_KEY, BaseDataFrameNormalizer
from xil._headers import get_url_response

_BOI_URL = "https://boi.org.il/PublicApi/GetExchangeRates"
_IDX = pd.MultiIndex.from_product(
    [[_CURRENCY_KEY], ["code", "official rate", "change (%)", "amount", "time"]]
)

get_boi_url_response = partial(get_url_response, set_context=True)


def get_boi_df(url: str = _BOI_URL) -> pd.DataFrame:
    """Get Bank of Israel (BOI) exchange rates"""
    with get_boi_url_response(url) as response:
        series = pd.read_json(response, typ="series", convert_dates=True)
    assert "exchangeRates" in series
    df = pd.json_normalize(series.exchangeRates)  # pylint: disable=no-member
    df.columns = _IDX
    df = BaseDataFrameNormalizer(df).norm()
    return df


if __name__ == "__main__":
    print(get_boi_df())



"""
Shared functionalities for retrieving URLs' data when headers are needed
"""

import http.client
import ssl
import urllib.request

USER_AGENT = "\
Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) \
AppleWebKit/537.36 (KHTML, like Gecko) \
Chrome/107.0.0.0 Safari/537.36"

UA_HEADER = {"User-Agent": USER_AGENT}

_DEFAULT_CONTEXT = ssl.create_default_context()
_DEFAULT_CONTEXT.set_ciphers("DEFAULT")


def get_url_response(
    url: str,
    set_context: bool = False,
) -> http.client.HTTPResponse:
    """
    Return the response from a URL with custom headers and SSL context when opening if
    set_context is True.
    """
    context = _DEFAULT_CONTEXT if set_context else None
    request = urllib.request.Request(url)
    return urllib.request.urlopen(  # type: ignore[no-any-return]
        request, context=context
    )



"""
Discount bank

https://www.discountbank.co.il/private/general-information/foreign-currency-transfers/exchange-rates/
Takes the data from:
https://www.discountbank.co.il/api/ExchangeRates/GetExchangeRates?bankCode=0011
"""

import pandas as pd

from xil._df_normalizer import BaseDataFrameNormalizer

_DISCOUNT_URL = "\
https://www.discountbank.co.il/api/ExchangeRates/GetExchangeRates?bankCode=0011"
_IDX0 = pd.MultiIndex.from_product([["currency"], ["code", "amount", "official rate"]])
_IDX1 = pd.MultiIndex.from_product([["cash", "transfer"], ["buy", "sell"]])
_DISCOUNT_IDX = _IDX0.append(_IDX1)


def get_discount_df(url: str = _DISCOUNT_URL) -> pd.DataFrame:
    """Get Discount Bank exchange rates"""
    s = pd.read_json(url, typ="series")
    df = pd.json_normalize(s.data["exchangeRates"]["CurrenciesList"])
    df.columns = _DISCOUNT_IDX
    df = BaseDataFrameNormalizer(df).norm()
    return df



"""
Massad bank

https://www.bankmassad.co.il/wps/portal/FibiMenu/Marketing/Private/ForeignCurrency/ForexOnline/Rates
Choose "Against Shekel" (״מול השקל״) tab

The data is directly taken from FIBI bank:
https://apps.fibi.co.il/Matach/matach.aspx

Use the `fibi` module instead.
"""



"""
Leumi's official exchange web page:
https://www.leumi.co.il/Lobby/currency_rates/40806/
embeds the table from:
https://www.bankleumi.co.il/vgnprod/currency/new_shaar_muskamim.html
https://www.bankleumi.co.il/Rates/Forex/Scripts/drawTable.js
which in turn consumes its data from:
https://www.bankleumi.co.il/Rates/Api3.5/ForeignExchange/LatestSelectedRates.aspx

Historical data by currency ID:
https://www.bankleumi.co.il/vgnprod/currency/ExchangeRateByCurrency.aspx?in_matbea=1

For business data:
https://biz.leumi.co.il/portal/site/Business/home_03/currency_rates/12283/
the table is embedded from the static page:
https://www.bankleumi.co.il/vgnprod/ltrade_new_shaar_muskamim_multilang_vgn_HE.html

Looks like the business data is identical to the private data.
"""

import pandas as pd

from xil._df_normalizer import JPYNormalizer
from xil._headers import UA_HEADER

_LEUMI_URL = "\
https://www.bankleumi.co.il/Rates/Api3.5/ForeignExchange/LatestSelectedRates.aspx"
_IDX0 = pd.MultiIndex.from_product(
    [["currency"], ["name", "official rate", "change (%)"]]
)
_IDX1 = pd.MultiIndex.from_product([["transfer", "cash"], ["buy", "sell"]])
_IDX = _IDX0.append(_IDX1)


class LeumiNormalizer(JPYNormalizer):
    """Leumi bank data normalizer"""

    _QUOT = "&quot;"

    @classmethod
    def _fix_quot(cls, raw_name: str) -> str:
        """fix " in 'דולר ארה&quot;ב'"""
        return raw_name.replace(cls._QUOT, '"')

    @classmethod
    def _preprocess_names(cls, names: pd.Series) -> pd.Series:
        names = super()._preprocess_names(names)
        return names.apply(cls._fix_quot)


def get_leumi_df(url: str = _LEUMI_URL) -> pd.DataFrame:
    """Get Leumi Bank exchange rates"""
    series = pd.read_json(url, typ="series", storage_options=UA_HEADER)
    # date = s.yatzigDate  # Hour in `s.topHeaderText`
    # pylint: disable-next=redefined-outer-name
    df = pd.json_normalize(series.data)
    df = df[
        [
            "currencyName",
            "yatzig",
            "percent",
            "hamchaot.knia",
            "hamchaot.mechira",
            "mezuman.knia",
            "mezuman.mechira",
        ]
    ]
    df.columns = _IDX
    df = df.loc[df[("currency", "name")] != "סל המטבעות", :]
    df = LeumiNormalizer(df).norm()
    return df


if __name__ == "__main__":
    df = get_leumi_df()
    print(df)



"""
Mizrahi-Tefahot Bank scraper foreign exchange rates data:
https://www.mizrahi-tefahot.co.il/brokerage/foreignexchange/
Exchange calculator:
https://www.mizrahi-tefahot.co.il/brokerage/currancyexchange/
"""

import pandas as pd

from xil._df_normalizer import DataFrameNormalizer

_MIZRAHI_TEFAHOT_URL = "https://www.mizrahi-tefahot.co.il/brokerage/foreignexchange/"
_IDX0 = pd.MultiIndex.from_product([["currency"], ["name", "amount", "official rate"]])
_IDX1 = pd.MultiIndex.from_product([["cash", "transfer"], ["buy", "sell"]])
_MIZRAHI_TEFAHOT_IDX = _IDX0.append(_IDX1)


def get_mizrahi_teafhot_df(url: str = _MIZRAHI_TEFAHOT_URL) -> pd.DataFrame:
    """Get Mizrahi Tefahot Bank exchange rates"""
    df = pd.read_html(url, header=0)[0]
    df.columns = _MIZRAHI_TEFAHOT_IDX
    df = DataFrameNormalizer(df).norm()
    return df


if __name__ == "__main__":
    print(get_mizrahi_teafhot_df())



"""
FIBI - First International Bank of Israel (HaBank HaBeinleumi)

The exchange rates page
https://www.fibi.co.il/wps/portal/FibiMenu/Marketing/Private/ForeignCurrency/Trade/Rates
includes static data from:
https://apps.fibi.co.il/Matach/matach.aspx
"""

import pandas as pd

from xil._df_normalizer import DataFrameNormalizer
from xil._headers import UA_HEADER

_FIBI_URL = "http://apps.fibi.co.il/Matach/matach.aspx"
_ENCODING = "iso-8859-8"
_MATCH = "העברות והמחאות"
_HEADER = [0, 1]
_ATTRS = {"class": "clsPart"}
_RELEVANT_COLS = [
    ("מטבע", "מטבע"),
    ("יחידה", "יחידה"),
    ("יציג", "יציג"),
    ("בנקנוטים", "מכירה"),
    ("בנקנוטים", "קניה"),
    ("העברות והמחאות", "מכירה"),
    ("העברות והמחאות", "קניה"),
]
_IDX0 = pd.MultiIndex.from_product([["currency"], ["name", "amount", "official rate"]])
_IDX1 = pd.MultiIndex.from_product([["cash", "transfer"], ["sell", "buy"]])
_IDX = _IDX0.append(_IDX1)


def get_fibi_df(url: str = _FIBI_URL) -> pd.DataFrame:
    """Get FIBI exchange rates"""
    dfs = pd.read_html(
        url,
        match=_MATCH,
        header=_HEADER,
        encoding=_ENCODING,
        attrs=_ATTRS,
        storage_options=UA_HEADER,
    )

    df = dfs[0]  # guaranteed to have at least one element at this point
    df = df[_RELEVANT_COLS]
    df.columns = _IDX
    df = DataFrameNormalizer(df).norm()
    return df


if __name__ == "__main__":
    print(get_fibi_df())



"""
Mercantile bank

https://www.mercantile.co.il/en/private/foregin-currency/exchange-rates/
Taken from:
https://www.mercantile.co.il/api/ExchangeRates/GetExchangeRates?bankCode=0017

The structure is identical Discount's, but the data is different.
"""

import pandas as pd

from xil.discount import get_discount_df as _get_discount_df

_MERCANTILE_URL = "\
https://www.mercantile.co.il/api/ExchangeRates/GetExchangeRates?bankCode=0017"


def get_mercantile_df(url: str = _MERCANTILE_URL) -> pd.DataFrame:
    """Get Mercantile Bank exchange rates"""
    # The structure is identical Discount's, but the data is different - use Discount's
    # function
    return _get_discount_df(url)
